/* 
 *  Name: sd.c
 *  Author: Mathäus Sander & Florian Weinhold
 *
 *  Desc.: This file contains the method bodies for the prototypes of sd.h.
 * 
 *  License: LGPL
*/

/* ################################################################## *
 * ##                                                              ## *
 * ##    Inlcudes                                                  ## *
 * ##                                                              ## *
 * ################################################################## */
 
#include "sd.h"

/* ################################################################## *
 * ##                                                              ## *
 * ##    Prototypes                                                ## *
 * ##                                                              ## *
 * ################################################################## */
 
//Reads the next 4 bytes after a command with R3 response
static uint32_t sd_read_four_byte(void);

/* ################################################################## *
 * ##                                                              ## *
 * ##    Variables                                                 ## *
 * ##                                                              ## *
 * ################################################################## */
 
/*
 * sd_status is a global variable to store the card status.
 * 
 * Bit 0: 1 = MMC
 * Bit 1: 1 = SD Card spec. V1
 * Bit 2: 1 = SD Card spec. V2
 * Bit 3: 1 = SDHC
 * Bit 4: 1 = Initialized, 0 = Not inititialized
 * Bit 5: 1 = Reading multiple blocks started
 * Bit 6: 1 = Writing multiple blocks started
 * 
 * If sd_status = 0, the card is not supported or has not been initialized.
 * */
volatile uint8_t sd_status = 0; 

/* ################################################################## *
 * ##                                                              ## *
 * ##    Methods - Initialization                                  ## *
 * ##                                                              ## *
 * ################################################################## */

/**
 * @brief sd_init initializes the SD card and collect information about it.
 * 
 * Procedure:
 * 
 * step 1. Set SPI Clock <400
 * step 2. Initialize Hardware
 * step 3. CMD0 to reach IDLE mode
 * step 4. CMD8 to check voltage range and get card version
 * step 5. ACMD41 to initialize Card
 * step 6. If 5. not work try CMD1 (for MMCs)
 * 
 * @return Card inforamtion
 * Bit 0: 1 = MMC
 * Bit 1: 1 = SD Card spec. V1
 * Bit 2: 1 = SD Card spec. V2
 * Bit 3: 1 = SDHC
 * Bit 4: 1 = Initialized, 0 = Not inititialized
 * Bit 5-7: reserved
 * */
uint8_t sd_init(void){
	
	//--------------------------------------Step 1 ---------------------------//
	
	//Set SPI Clock < 400kHz
	spi_speed_low();
	
	
	//--------------------------------------Step 2 ---------------------------//
	
	/* Hardware initialisation */
	
	//Chips select for SDCard to output
    SD_CS_DDR |= ( 1 << SD_CS_PIN );	
    //CS to logic 1
    sd_cs_high();
    
    //send 74+ dummy clocks
    for(uint8_t i = 0; i<8 ; i++)
        spi_transmit_byte(SD_DUMMY_BYTE);
	
	
	//--------------------------------------Step 3 ---------------------------//
	
	#if defined(DEBUG) && defined(DEBUG_INIT)
		printf("Go IDLE mode ... \n");
	#endif
	
	uint8_t timeout = SD_INIT_TIMEOUT;
	uint8_t rec = 0;
	
    //CMD0 to get Card into IDLE State
    while(rec != SD_R1_IDLE_STATE){
		rec = sd_send_cmd( SD_GO_IDLE_STATE, 0 );
		if(!timeout--){
			#if defined(DEBUG) && defined(DEBUG_INIT)
				printf("Timeout reached. Stop card initialisation\n");
			#endif
			return  sd_status = SD_TYPE_UNSUPPORTED;
		}
	}
	
		
    #if defined(DEBUG) && defined(DEBUG_INIT)
		printf("Card is in IDLE mode. Begin to verify version\n");
	#endif

	
	//--------------------------------------Step 4 ---------------------------//
	/*
	 * CMD8 is used to check Voltage and expend new functionality to some CMDs
	 * CMD41 is expand to support Initialisation of SDHC and SDXC Cards.
	 * */
	 
	//send CMD08
	rec = sd_send_cmd(SD_SEND_IF_COND, 0x1AA);
	
	if( (rec & SD_R1_ILLEGAL_COMMAND)){ //means MMC/SDC V1
		 sd_status |= SD_TYPE_SPEC1;
		#if defined(DEBUG) && defined(DEBUG_INIT)
			printf("MMC/SDCV1\n");
		#endif
	}else{ //SDC V2 or SDHC
		uint32_t ocr = sd_read_four_byte();
		
		if((uint8_t)((ocr>>8)&0xFF) != 1){
			//Voltage error
			#if defined(DEBUG) && defined(DEBUG_INIT)
				printf("Voltage error\n");
			#endif
			//return  sd_status = SD_TYPE_UNSUPPORTED;
		}
		
		if((uint8_t)(ocr&0xFF) != 0xAA){
			//Pattern error
			#if defined(DEBUG) && defined(DEBUG_INIT)
				printf("Pattern error\n");
			#endif
			//return  sd_status = SD_TYPE_UNSUPPORTED;
		}
	}
	
	
	//--------------------------------------Step 5 ---------------------------//
	
	
	/* Software initialisation - Initialisation with ACMD41 */
	
	timeout = SD_INIT_TIMEOUT;
	while(rec){
		sd_send_cmd(SD_APP_CMD, SD_HC); //CMD55
		rec = sd_send_cmd(SD_SD_SEND_OP_COND, SD_HC);//ACMD41
		if(!(--timeout)){
			#if defined(DEBUG) && defined(DEBUG_INIT)
				printf("Timeout reached. Stop card initialisation with ACMD41\n");
			#endif
			break;
		}
	}
	
	
	//--------------------------------------Step 6 ---------------------------//
	
	/* Software initialisation - Initialisation with CMD1 */
	#if defined(DEBUG) && defined(DEBUG_INIT)
		if(rec)
			printf("Try to initiate with CMD1\n");
	#endif
	timeout = SD_INIT_TIMEOUT;
	while(rec){
		rec = sd_send_cmd(SD_SEND_OP_COND, 0); //CMD1
		if(!(--timeout)){
			#if defined(DEBUG) && defined(DEBUG_INIT)
				printf("Timeout reached. Stop card initialisation\n");
			#endif
			return sd_status = SD_TYPE_UNSUPPORTED;
		}
	}
	
	spi_speed_high();
	#if defined(DEBUG) && defined(DEBUG_INIT)
		printf("Card init completed. Now the Card can read / write.\n");
	#endif
	return rec;
	
} //  sd_init(void)  

//______________________________________________________________________

/* ################################################################## *
 * ##                                                              ## *
 * ##    Methods - Reading                                         ## *
 * ##                                                              ## *
 * ################################################################## */
 
 /**
 * @brief Reads the next four bytes on SPI-BUS. (Useful to get OCR after a command)
 * @return 4 Byte long value, read by SPI
 * */
static uint32_t sd_read_four_byte(){
	
	uint32_t values = (uint32_t)spi_read_byte()<<24;
	values |= (uint32_t)spi_read_byte()<<16;
	values |= (uint32_t)spi_read_byte()<<8;
	values |= spi_read_byte();
	
	#ifdef DEBUG
		printf("ocr: 0x%x 0x%x 0x%x 0x%x\n", (uint8_t)((values>>24)&0xFF), (uint8_t)((values>>16)&0xFF), (uint8_t)((values>>8)&0xFF), (uint8_t)(values&0xFF));
	#endif	
	return values;
	return 0;
} // sd_read_four_byte()

//______________________________________________________________________

/**
 * @brief sd_read_block reads a single 512 byte block of data from the SD card and write it into the buffer.
 * 
 * read precoedure:
 * Step 1. Send CMD17 to signal Card that a single Block should be read.
 * Step 2. Check for errors
 * Step 3. Wait for start token 0xfe
 * Step 4. Read bytes and write into buffer
 * Step 5. read dummy CRC (can be ignored under SPI)
 * 
 * @param sector   - number of sector.
 * @param *buffer  - Pointer to the buffer
 * @param buf_size - Size of the buffer
 * @return 
 * */
uint8_t sd_read_single_block(uint32_t sector, uint8_t *buffer, uint16_t buf_size){
	
	//--------------------------------------Step 1 ---------------------------//
	
	//Send cmd17 to MMC/SDC/SDHC Card
	uint8_t rec = sd_send_cmd(SD_READ_SINGLE_BLOCK, sector<<9); //read a Block command
	
	//--------------------------------------Step 2 ---------------------------//
	
	//Check for errors 
	if(rec != 0x00) 
		return rec; 
	
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("CMD to read accepted ! Card starts to write Data into internal buffer.\n");
	#endif 
	
	/*read Bytes and write to Buffer*/
	sd_cs_low();
	
	//--------------------------------------Step 3 ---------------------------//
	
	//Wait for startbyte 
	uint8_t timeout = SD_READ_TIMEOUT;
	while( (rec = spi_read_byte()) != SD_START_TOCKEN){
		
		#if defined(DEBUG) && defined(DEBUG_READ)
			printf("rec = %x ",rec);
		#endif 
		if(!(timeout--)){
			#if defined(DEBUG) && defined(DEBUG_READ)
				printf("read-Timeout reached\n");
			#endif 
			sd_cs_high();
			return -1;
		} 
	}
	
	
	//--------------------------------------Step 4 ---------------------------//
	
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("rec=0xfe. Card is now rdy to send bytes.\n");
		printf("Start reading %d bytes\n", buf_size);
	#endif
	
	//read buf_size bytes
	
	for(uint16_t i=0; buf_size; buf_size--, i++){ 
		*(buffer++) = spi_read_byte();
	}
	
	//--------------------------------------Step 5 ---------------------------//
	
	//receive CRC
	spi_read_byte(); //ignored
	spi_read_byte(); //ignored
	
	sd_cs_high(); 
	
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("finished reading.\n");
	#endif
	return 0;
} //  sd_read_single_block(uint32_t, uint8_t *, uint16_t){

//______________________________________________________________________

/**
 * @brief sd_start_reading_multiple_block starts the process of reading 
 * multiple blocks, starting at the given sector
 * 
 * precoedure:
 * 1. 
 * 
 * @param sector - number of sector in wich the reading process begin
 * @return returns the last recieved byte from the Card
 * */
uint8_t sd_start_reading_multiple_block(uint32_t sector){
	
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("Starts multiple Block reading at Sektor no. 0x%x %x %x %x\n", (uint8_t)((sector>>24)&0xFF), (uint8_t)((sector>>16)&0xFF), (uint8_t)((sector>>8)&0xFF), (uint8_t)(sector&0xFF));
	#endif 
	
	//If no SDHC Card is inserted, the sector number is multiplized with 512
	//if(!(sd_status&SD_TYPE_SDHC)) 
		sector = sector << 9; // *512 
	
	uint8_t rec = sd_send_cmd(SD_READ_MULTIPLE_BLOCK, sector);
	
	if(rec == SD_R1_ILLEGAL_COMMAND)
		return rec;
		
	uint8_t timeout = SD_READ_MULTI_TIMEOUT;
	
	while ((rec = spi_read_byte()) != SD_START_TOCKEN){
		#if defined(DEBUG) && defined(DEBUG_READ)
		printf("Timeout = %d und rec=%x\n", timeout, rec);
		#endif 
		if(!(timeout--)){
			#if defined(DEBUG) && defined(DEBUG_READ)
				printf("timeout was reached while waiting for the start token\n");
			#endif 
			return rec;
		}
	};
	
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("MBR is now enable.\n");
	#endif 
	return rec;
	
} //sd_start_reading_multiple_block(uint32_t)

//______________________________________________________________________

uint8_t sd_read_next_block(uint8_t *buffer, uint16_t buf_size){
	#if defined(DEBUG) && defined(DEBUG_READ)
		printf("Read next Block.\n");
	#endif 
	for(int i = 0; i < buf_size; i++){
		*buffer++ = spi_read_byte();
	}

	spi_read_byte();	// CRC can be ignored in SPI mode.
	spi_read_byte();	// CRC can be ignored in SPI mode.
	
	uint8_t rec = 0;
	uint8_t timeout = SD_READ_MULTI_TIMEOUT;
	
	while ((rec = spi_read_byte()) != SD_START_TOCKEN){
		if(!(timeout--)){
			#if defined(DEBUG) && defined(DEBUG_READ)
				printf("timeout was reached while waiting for the start token\n");
			#endif 
			return rec;
		}
	};
	
	return 0;
	
} //sd_read_next_block(uint8_t *, uint16_t)

//______________________________________________________________________


uint8_t sd_stop_reading_multiple_block(void){
	
	uint8_t rec = spi_transmit_byte(SD_STOP_TRANSMISSION);		
	
	spi_read_byte(); // dummy byte nach cmd12

	return rec;
	
} //sd_stop_reading_multiple_block(void)

//______________________________________________________________________

 /*################################################################## *
 * ##                                                              ## *
 * ##    Methods - Writing                                         ## *
 * ##                                                              ## *
 * ################################################################## */
 
 
/**
 * @brief sd_write_block writes a single 512 byte block of data to the SD card.
 * 
 * write precoedure:
 * 1. Send CMD24 to signal Card that a single block(512 byte) of data will be send.
 * 2. Check for errors
 * 3. send start token 0xfe
 * 4. write Data to SPI
 * 5. send dummy CRC
 * 6. check response token
 * 
 * @param address - SDSC Card uses byte unit address and SDHC and SDXC Cards use block unit address (512 bytes unit).
 * @param *buffer - Pointer to the buffer
 * @param buf_size - Size of the buffer
 * @return 0 = all ok, -1 = error 
 * */
uint8_t sd_write_single_block(uint32_t address, uint8_t *buffer, uint16_t buf_size){
	
	//--------------------------------------Step 1 ---------------------------//
	
	//Send cmd24 to MMC/SDC/SDHC Card
	uint8_t rec = sd_send_cmd(SD_WRITE_SINGLE_BLOCK, address<<9);
	
	//--------------------------------------Step 2 ---------------------------//
	
	//Check for errors 
	if(rec != 0x00) 
		return rec; 
	
	#if defined(DEBUG) && defined(DEBUG_WRITE)
		printf("CMD to write accepted !\n");
		printf("send start token\n");
	#endif 
	
	//--------------------------------------Step 3 ---------------------------//
	
	//send start token
	spi_transmit_byte(SD_START_TOCKEN);	
	
	#if defined(DEBUG) && defined(DEBUG_WRITE)
		printf("send Data\n");
	#endif 
	
	//--------------------------------------Step 4 ---------------------------//
	
	//write bytes from buffer to spi
	for (uint16_t i=0; i<512; i++) {
		spi_transmit_byte(*(buffer++));
	}
	
	//--------------------------------------Step 5 ---------------------------//
	
	//send dummy CRC-Bytes
	spi_transmit_byte(SD_DUMMY_BYTE); 
	spi_transmit_byte(SD_DUMMY_BYTE); 
	
	//--------------------------------------Step 6 ---------------------------//
	
	//check response for token XXX00101
	rec = spi_transmit_byte(SD_DUMMY_BYTE);
	if((rec&0x1F) != 0x05){
		return rec;
	}

	while (spi_transmit_byte(SD_DUMMY_BYTE) != 0xff){};
	
	sd_cs_high();
	#if defined(DEBUG) && defined(DEBUG_WRITE)
		printf("Write process finished.\n");
	#endif
	return 0;
} //  sd_write_single_block(uint32_t, uint8_t *, uint16_t){

 //______________________________________________________________________
 
 uint8_t sd_start_writing_multiple_block(uint32_t address){

	return sd_send_cmd(SD_WRITE_MULTIPLE_BLOCK, address);
	
} //sd_start_writing_multiple_block(uint32_t)

//______________________________________________________________________

uint8_t sd_write_next_block(uint8_t *buffer, uint16_t buf_size){
	return 0;
} //sd_write_next_block(uint8_t *, uint16_t)

//______________________________________________________________________


void sd_stop_writing_multiple_block(void){
	
	//send token
	spi_transmit_byte(SD_MBWRT_STOP_TOCKEN);	
	
	
} //sd_stop_writing_multiple_block(void)

//______________________________________________________________________

 /*################################################################## *
 * ##                                                              ## *
 * ##    Methods - Debug	                                       ## *
 * ##                                                              ## *
 * ################################################################## */
 
 /**
 * @brief Sends a formated debug message via UART.
 * @param *buffer - Buffer to print
 * @param buf_size - size of the Buffer
 * */
void sd_print_buffer(uint8_t *buffer, uint16_t buf_size){
	#ifdef DEBUG
	
	printf("offset\t0\t1\t2\t3\t4\t5\t6\t7\t8\t9\tA\tB\tC\tD\tE\tF\n");
	uint8_t offset = 0;
	for(uint16_t i = 0; i<buf_size; i=i+16){
		printf("0x%x", offset++);  //ffset
		printf("\t%x", *buffer++); //0
		printf("\t%x", *buffer++); //1
		printf("\t%x", *buffer++); //2
		printf("\t%x", *buffer++); //3
		printf("\t%x", *buffer++); //4
		printf("\t%x", *buffer++); //5
		printf("\t%x", *buffer++); //6
		printf("\t%x", *buffer++); //7
		printf("\t%x", *buffer++); //8
		printf("\t%x", *buffer++); //9
		printf("\t%x", *buffer++); //A
		printf("\t%x", *buffer++); //B
		printf("\t%x", *buffer++); //C
		printf("\t%x", *buffer++); //D
		printf("\t%x", *buffer++); //E
		printf("\t%x", *buffer++); //F
		printf("\n"); 
	}
	
	#endif	
} //  sd_print_buffer(uint8_t *, uint16_t)

//______________________________________________________________________

 /*################################################################## *
 * ##                                                              ## *
 * ##    Methods - Other                                           ## *
 * ##                                                              ## *
 * ################################################################## */

/**
 * @brief sd_send_cmd helps to send a command to the card. A command is always 6 bytes long 
 * [start bit ('0') + transmission bit('1') + command index (6bit) + arguments (4 byte) + CRC7 (7bit)+ end bit('1')].
 * 
 * precoedure:
 * 1. build CMD index + Argument
 * 2. add CRC
 * 3. send CMD
 * 4. recieve answer
 * 
 * @param cmd - six bits long command index of the Command (for example 000101 for CMD5)
 * @param arg - four byte long argument 
 * 
 * @return response of the Command (Format R1)
 * */
uint8_t sd_send_cmd(uint8_t cmd, uint32_t arg){
	
	//--------------------------------------Step 1 ---------------------------//
	
	/*build complete Command*/
	uint8_t buffer[6] ;                   	//temporary Buffer for the CMD
	buffer[0] = cmd | SD_CMD_MASK;        	//Set start and transmission bit + Command
	buffer[1] = (arg & 0xFF000000) >> 24; 	//Set parameter byte 1
	buffer[2] = (arg & 0x00FF0000) >> 16; 	//Set parameter byte 2 
	buffer[3] = (arg & 0x0000FF00) >> 8;  	//Set parameter byte 3
	buffer[4] = (arg & 0x000000FF);       	//Set parameter byte 4
	
	//--------------------------------------Step 2 ---------------------------//
	
	//cyclic redundancy check (dont care in SPI mode unless CMD0 (0x95) and CMD8 (0x87)
	if(cmd == SD_GO_IDLE_STATE)
		buffer[5] = 0x95;               	
	else if(cmd == SD_SEND_IF_COND)
		buffer[5] = 0x87;
	else
		buffer[5] = 0xFF;
		
	#if defined(DEBUG) && defined(DEBUG_CMD)
		//Prints the Command
		printf("Sending CMD: ");
		printf("%d, ", buffer[0]&0b00111111);
		printf("ARGS: ");
		for(uint8_t i=1; i<5; i++)
			printf("0x%x, ", buffer[i]);
		printf("CRC: 0x%x, ", buffer[5]);
	#endif
	
	//--------------------------------------Step 3 ---------------------------//
	
	sd_cs_low();
	
	//Send Dummybyte (timing reasons)
	spi_transmit_byte(SD_DUMMY_BYTE); 
	
	//Send complete Command via SPI
    for(uint8_t i=0; i<6; i++)
        spi_transmit_byte(buffer[i]); 
	
	//--------------------------------------Step 4 ---------------------------//
	
	//receive answer byte
	uint8_t rec = 0xFF;
	uint8_t timeout = SD_CMD_TIMEOUT;
	while(rec==0xFF){
		rec = spi_read_byte();
		if(!(timeout--)){
			#if defined(DEBUG) && defined(DEBUG_CMD)
				printf("CMD timeout reached. Last recieve = %x. \n", rec);
			#endif
			sd_cs_high();
			return rec;
		}
	}
	
	sd_cs_high();            //clear SPI	
	#if defined(DEBUG) && defined(DEBUG_CMD)
		//Prints the Command
		printf("recv.: 0x%x", rec);
		printf(".\n");
	#endif	
	
    return rec;
} //  sd_send_cmd(uint8_t, uint32_t)

//______________________________________________________________________

