/* 
 *  Name: sd.h
 *  Author: Mathäus Sander & Florian Weinhold
 *
 *  Desc.: This file contains various enable includes, defines and prototypes, which setting to SD and SPI communication.
 * 		   For example, methods to set the SPI clock, or to indicate which method should be called when a byte should be
 * 		   transmitted over SPI.
 * 
 */

#ifndef __SDCARD__
#define __SDCARD__

/* ################################################################## *
 * ##                                                              ## *
 * ##    Includes                                                  ## *
 * ##                                                              ## *
 * ##################################################################*/
 
#include "../uart/uart.h"
#include "../spi/spi.h"
#include <avr/io.h>
#include <util/delay.h>


/* ################################################################## *
 * ##                                                              ## *
 * ##    Options								                   ## *
 * ##                                                              ## *
 * ##    Note: Customize Defines for your need					   ## *
 * ##    										                   ## *
 * ################################################################## */
 
 
 /*-----------------------------Debugging--------------------------------
 * 
 * DEBUG controls the production of code for debugging functions. If it is defined, the debugging code is compiled.
 * It should be defined only when messages are required because otherwise Rescources wasted on unnecessary surgery.
 * */
#ifndef DEBUG
	#define DEBUG 
#endif
/*
 * If DEBUG_CMD and DEBUG are defined, debugging for Commands is enabled
 * */
#ifndef DEBUG_CMD
	#define DEBUG_CMD
#endif

/*
 * If DEBUG_INIT and DEBUG are defined, debugging for Card initialisation is enabled
 * */
#ifndef DEBUG_INIT
	#define DEBUG_INIT
#endif

/*
 * If DEBUG_WRITE and DEBUG are defined, debugging for write process is enabled
 * */
#ifndef DEBUG_WRITE
	#define DEBUG_WRITE
#endif

/*
 * If DEBUG_READ and DEBUG are defined, debugging for read process is enabled
 * */
#ifndef DEBUG_READ
	#define DEBUG_READ
#endif


/*-----------------------------Timeouts---------------------------------
 * 
 * SD_INIT_TIMEOUT is the number of loop executions in which a command is sent to initialize the card.
 * */
#define SD_INIT_TIMEOUT 		100 //value between 0 and 254 

/*
 * SD_READ_TIMEOUT is the number of loop executions in which a command is sent to read from Card.
 * */
#define SD_READ_TIMEOUT 		100 //value between 0 and 254 

/*
 * SD_READ_MULTI_TIMEOUT is the number of loop executions in which the start token should be recieved.
 * */
#define SD_READ_MULTI_TIMEOUT 	20 //value between 0 and 254 

/*
 * SD_WRITE_TIMEOUT is the number of loop executions in which a command is sent to write to the card.
 * */
#define SD_WRITE_TIMEOUT 		100 //value between 0 and 254 

/*
 * SD_CMD_TIMEOUT is the number of loop executions in which a command response should be recieved.
 * */
#define SD_CMD_TIMEOUT 			100 //value between 0 and 254 




/*-----------------------------SPI-Software-----------------------------
 * 
 * The following macro defines a method that is used to transmit and recieve a byte via SPI to the card.
 * Enter the appropriate method to use to SPI.
 * */
#define spi_transmit_byte(x) 	spi2_transmit(x) 

/*
 * The following macro defines a method that is used to read a byte via SPI from the card.
 * */
#define spi_read_byte() 		spi2_transmit(SD_DUMMY_BYTE) 



/*-----------------------------SPI-Hardware-----------------------------
 * 
 * The following macro sets the SPI clock to a frequency between 100kHz and 400kHz
 * */
#define spi_speed_low() 	{SPSR|=(1<<SPI2X);SPCR|=(1<<SPR1);SPCR&=~(1<<SPR0);} 	

/*
 * The following macro sets the SPI clock to the frequency with which the data should be transferred to the card.
 * */					
#define spi_speed_high()    {SPSR&=~(1<<SPI2X);SPCR&=~(1<<SPR1);SPCR&=~(1<<SPR0);} 	

/*defines for Hardware Initialisation*/
#define SD_CS_PIN  			PC6    //Chipselect for SD Card
#define SD_CS_PORT 			PORTC  //Port of SD_CS_PIN
#define SD_CS_DDR  			DDRC   //Data Direction Register of SD_CS_PIN

/*
 * The following macro is defined for the SD card to set the previously defined chip select to a logical 1.
 * */
#define sd_cs_low()   		SD_CS_PORT &= ~(1 << SD_CS_PIN) // SD_CS -> '1'

/*
 * The following macro is defined for the SD card to set the previously defined chip select to a logical 0.
 * */
#define sd_cs_high()  		SD_CS_PORT |= (1 << SD_CS_PIN)  // SD_CS -> '0'



/* ################################################################## *
 * ##                                                              ## *
 * ##    Defines - SD Software specific                            ## *
 * ##                                                              ## *
 * ##################################################################*/
 
/*
 * SD_DUMMY_BYTE is the value that is sent when it should be read via SPI only.
 * */ 
#define SD_DUMMY_BYTE 				0xFF//

/*-----------------------------SD-Status defines------------------------------
 * 
 * The SD Status in this software is save in a one byte [bit 0...7] long value.
 * 
 * Bit 0: 1 = MMC
 * Bit 1: 1 = SD Card spec. V1
 * Bit 2: 1 = SD Card spec. V2
 * Bit 3: 1 = SDHC
 * Bit 4: 1 = Initialized, 0 = Not inititialized
 * Bit 5-7: reserved
 * 
 * */ 
#define SD_DUMMY_BYTE 				0xFF//
#define SD_TYPE_UNSUPPORTED			0 
#define SD_TYPE_MMC					(1<<0)
#define SD_TYPE_SPEC1				(1<<1)
#define SD_TYPE_SPEC2 				(1<<2)	 
#define SD_TYPE_SDHC 				(1<<3)

/*Masks to check SD Status*/

#define SD_STATUS_INITIALIZED      	(1<<4)

#define SD_READ_MULTIPLE_ENABLED   	(1<<5)

#define SD_WRITE_MULTIPLE_ENABLED  	(1<<6)

#define SD_STATUS_MASK 				0x1F

/*
 *	SD_HC is used for CMDs who needs to set the HC byte (e.g. ACMD41)
 * */ 
#define SD_HC 						(1L<<30)

/*-----------------------------SD-Commands------------------------------
 * 
 * Command Format (6 Bytes): 
 * start bit ('0') + transmission bit('1') + command index (6bit) + arguments (4 byte) + CRC7 (7bit)+ end bit('1')
 * 
 * Currently, not all CMDs used.
 * */
 
										 //	Command	| Return	| Arguments (stuff bits should be 0)
										 //	--------+-----------+-------------------------------------------------------------------------
#define SD_GO_IDLE_STATE  			0x00 //	CMD0 	| R1		| [31:0] Stuff bits
#define SD_SEND_OP_COND  			0x01 //	CMD1	| R1		| [31]Reserved bit [30]HCS [29:0]Reserved bits
#define SD_SWITCH_FUNC				0x06 //	CMD6	| R1		| see Specification
#define SD_SEND_IF_COND				0x08 //	CMD8	| R7		| [31:12]Reserved bits [11:8]supply voltage voltage(VHS) [7:0]check pattern
#define SD_SEND_CSD 				0x09 //	CMD9	| R1		| [31:0] Stuff bits
#define SD_SEND CID 				0x0A //	CMD10	| R1		| [31:0] Stuff bits
#define SD_STOP_TRANSMISSION 		0x0C //	CMD12	| R1b		| [31:0] Stuff bits
#define SD_SEND_STATUS 				0x0D //	CMD13	| R1		| [31:0] Stuff bits
#define SD_SET_BLOCKLEN 			0x10 //	CMD16	| R1		| [31:0] Block length
#define SD_READ_SINGLE_BLOCK 		0x11 //	CMD17	| R1		| [31:0] Data address(*)
#define SD_READ_MULTIPLE_BLOCK 		0x12 //	CMD18	| R1		| [31:0] Data address(*)
#define SD_WRITE_SINGLE_BLOCK		0x18 //	CMD24	| R1		| [31:0] Data address(*)
#define SD_WRITE_MULTIPLE_BLOCK		0x19 //	CMD25	| R1		| [31:0] Data address(*)
#define SD_PROGRAM_CSD 				0x1B //	CMD27	| R1		| [31:0] Stuff bits
#define SD_SET_WRITE_PROT  			0x1C //	CMD28	| R1b		| [31:0] Data address(*)
#define SD_CLR_WRITE_PROT 			0x1D //	CMD29	| R1b		| [31:0] Data address(*)
#define SD_SEND_WRITE_PROT 			0x1E //	CMD30	| R1		| [31:0] Write protect data address(*)
#define SD_ERASE_WR_BLK_START_ADDR 	0x20 //	CMD32	| R1 		| [31:0] Data address(*)
#define SD_ERASE_WR_BLK_END_ADDR 	0x21 //	CMD33	| R1		| [31:0] Data address(*)
#define SD_ERASE					0x26 //	CMD38	| R1b		| [31:0] Stuff bits
#define SD_LOCK_UNLOCK 				0x2A //	CMD42	| R1		| [31:0] Reserved bits set all to 0
#define SD_APP_CMD 					0x37 //	CMD55	| R1		| [31:0] Stuff bits
#define SD_GEN_CMD 					0x38 //	CMD56	| R1		| [31:1] Stuff bits [0] RD/WR
#define SD_READ_OCR 				0x3A //	CMD58	| R3		| [31:0] Stuff bits
#define SD_CRC_ON_OFF 				0x3B //	CMD59	| R1		| [31:1] Stuff bits [0] CRC option

#define SD_SD_SEND_OP_COND			0x29 //	ACMD41	| R1		| [31]Resetved bit [30]HCS [29:0]Reserved bits

// (*) SDSC Card (CCS=0) uses byte unit address and SDHC and (CCS=1) use block unit address

/*
 * SD_CMD_MASK is a mask for the first two bits (start and transmission) of a command. For this, the mask 
 * and the command index should be linked with an OR (e.g. (SD_SEND_CSD | SD_CMD_MASK))
 **/
#define SD_CMD_MASK 				0x40; //star bit and transmission bit



/*-------------------------SD-Command Response---------------------------
 * 
 * Below are definitions for the return values ​​of commands. Depending on the command R1, R1b, R3 or R7 is returned.
 * 
 * Format R1 (1 byte)
 * Format R1b (like R1 with busy waiting)
 * Format R3 (R1 + 4 Byte OCR)
 **/

#define SD_R1_IDLE_STATE 			(1<<0) // 
#define SD_R1_ERASE_RESET  			(1<<1) //
#define SD_R1_ILLEGAL_COMMAND 		(1<<2) //
#define SD_R1_COM_CRC_ERR 			(1<<3) //
#define SD_R1_ERASE_SEQ_ERR 		(1<<4) //
#define SD_R1_ADDRESS_ERR			(1<<5) //
#define SD_R1_PARAMETER_ERR 		(1<<6) //


#define SD_R3_IDLE_STATE 			(1<<0) //
#define SD_R3_ERASE_RESET 			(1<<1) //
#define SD_R3_ILLEGAL_COMMAND 		(1<<2) //
#define SD_R3_COM_CRC_ERR 			(1<<3) //
#define SD_R3_ERASE_SEQ_ERR 		(1<<4) //
#define SD_R3_ADDRESS_ERR 			(1<<5) //
#define SD_R3_PARAM_ERR 			(1<<6) //


/*-------------------------Data Response Tokens-------------------------
 * 
 * Every date block written to the card will be acknoweledged by a data
 * response token. (1 Byte, Format: x+x+x+0+Status+1)
 * 
 * DRT := Data Response Token
 * CRC  = Cyclic redundancy check
 * WRT := Write
 */
 
 #define SD_DRT_DATA_ACCEPTED		0x05 	//Data Accepted
 #define SD_DRT_CRC_ERROR			0x0B	//Date rejected tue to a CRC Error
 #define SD_DRT_WRT_ERROR			0x0D	//Date rejected tue to a Write Error
 #define SD_DRT_MASK				0x1F 	//Mask for Bit 4...0
 
/*-------------------------Data Response Tokens ------------------------
 * 
 * Every date block written to the card will be acknoweledged by a data
 * response token. (1 Byte)
 * 
 * MBWRT := Multiple Block Write
 */

/*
 * SD_START_BLOCK_TOCKEN symbolizes the value, which is sent or received before a data transfer begins.
 * It is used for Single Block Read, Single Block Write and Multiple Block Read
 * */
#define SD_START_TOCKEN				0xFE// 

/*
 * SD_STOP_TOCKEN symbolizes the value, which is sent to stop a transfer.
 * IOt is used for Multiplse Block Write
 * */
#define SD_MBWRT_START_TOCKEN		0xFC//

/*
 * SD_STOP_TOCKEN symbolizes the value, which is sent to start a transfer.
 * IOt is used for Multiplse Block Write
 * */
#define SD_MBWRT_STOP_TOCKEN		0xFD//


	
/* ################################################################## *
 * ##                                                              ## *
 * ##    Prototypes                                                ## *
 * ##                                                              ## *
 * ################################################################## */

/*-------------------------Initialization---------------------------
 * */
 
//initialize SD Card 
uint8_t sd_init(void);

/*-------------------------Reading----------------------------------
 * */
 
//Reads a block from the MMC/SD/SDHC Card
uint8_t sd_read_single_block(uint32_t, uint8_t *, uint16_t);

//Starts the process of reading multiple blocks, starting at the given address
uint8_t sd_start_reading_multiple_block(uint32_t);

//Reads the next block after sd_start_reading_multiple_block was called.
uint8_t sd_read_next_block(uint8_t *, uint16_t);

//Stops the process of reading multiple blocks
uint8_t sd_stop_reading_multiple_block(void);

/*-------------------------Writing-----------------------------------
 * */
 
//Writes a block to the MMC/SD/SDHC Card
uint8_t sd_write_single_block(uint32_t, uint8_t *, uint16_t);

//Starts the process of writing multiple blocks, starting at the given address
uint8_t sd_start_writing_multiple_block(uint32_t);

//Writes the next block after sd_start_writing_multiple_block was called.
uint8_t sd_write_next_block(uint8_t *, uint16_t);

//Stops the procecss of writing multiple blocks
void sd_stop_writing_multiple_block(void);

/*-------------------------Debug-------------------------------------
* */

//Method to get a formated debug message (only if Debugging is enabled)
void sd_print_buffer(uint8_t *buffer, uint16_t);

/*-------------------------Other-------------------------------------
* */

//Sends a 6 Byte long command via SPI
uint8_t sd_send_cmd(uint8_t, uint32_t); 


	
/* ################################################################## *
 * ##                                                              ## *
 * ##  some defines for my own well-being when i read the code     ## *
 * ##                                                              ## *
 * ################################################################## */
 
#ifndef true
	#define true 1
#endif

#ifndef TRUE
	#define TRUE 1
#endif

#ifndef false
	#define false 0
#endif

#ifndef FALSE
	#define FALSE 0
#endif

#ifndef null
	#define null 0
#endif

#ifndef NULL
	#define NULL 0
#endif

#endif 
